/*****************************************************************************
 * channel.c: channel for Bt848 frame grabber driver
 *****************************************************************************
 * $Id: channel.c,v 1.5 2003/07/10 05:51:30 sight4 Exp $
 *****************************************************************************
 * Copyright (C) 2001 Keuleu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************
 *
 *  Original code:
 *  channel for Bt848 frame grabber driver
 *
 *  Copyright (C) 1996,97 Marcus Metzler (mocm@thp.uni-koeln.de)
 *
 *  many changes by Gerd Knorr <kraxel@cs.tu-berlin.de>
 *      [ hmm, think by now nearly nothing left from the original code ... ]
 *
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <X11/Intrinsic.h>
#ifndef NO_X11
# include <X11/StringDefs.h>
# include <X11/Xaw/XawInit.h>
# include <X11/Xaw/Command.h>
#endif

#include "config.h"
#include "strtab.h"
#include "grab.h"
#include "x11.h"
#include "channels.h"
#include "channel.h"
#include "mixer.h"
#include "plugin.h"
#include "memcpy.h"
#include "deinterlace.h"
#include "divx_ui.h"
#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
#include "divx.h"
#endif

/* ----------------------------------------------------------------------- */
/* misc common stuff, not only channel related                             */

struct CHANNEL defaults = { 
  "defaults",      // name
  NULL,            // key
  "5",             // cname
  0,               // channel
  0,               // fine
  0,               // freq
  CAPTURE_OVERLAY, // capture
  0,               // source
  0,               // norm
  32768,           // color
  32768,           // bright
  32768,           // hue
  32768,           // contrast
  0,               // pixmap
  0,               // button
  0,               // ckey
  DEINTERLACE_NONE // deinterlace
};

struct CHANNEL **channels = NULL;

int count = 0;
int have_mixer = 0;

int cur_sender = -1, cur_channel = 5, cur_fine = 0;
int cur_norm = -1, cur_input = -1;

int chan_tab = 4;

extern PLUGIN *plugin;
extern char user_dir[255];
extern int have_config;
extern struct GRABBER *grabbers[];
extern int grabber;
extern int fs_width, fs_height;
extern int fs_xoff, fs_yoff;
extern int pix_width, pix_height;
extern int owidth, oheight;
extern int wwidth, wheight;
extern int cur_color, cur_bright, cur_hue, cur_contrast, cur_capture,
  cur_volume;
extern struct STRTAB deinterlace_list[];
extern int deinterlace_method;
extern void set_deinterlace(int j);

#ifndef NO_X11
extern Display *dpy;
void button_cb (Widget widget, XtPointer clientdata, XtPointer call_data);

extern Widget app_shell, chan_box, chan_viewport, tv, opt_paned, c_stayontop, c_decoration;
#endif

#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
extern divx_t divx;
#endif
extern char *mixer;

extern int memcpy_method;
struct STRTAB memcpy_list[] = {
  {MEMCPY_PROBE,  "probe"},
  {MEMCPY_GLIBC,  "glibc"},
  {MEMCPY_KERNEL, "kernel"},
  {MEMCPY_MMX,    "mmx"},
  {MEMCPY_MMXEXT, "mmxext"},
  {MEMCPY_SSE,    "sse"},
  {-1, NULL}
};

/* ----------------------------------------------------------------------- */
void save_config (void);
/* ----------------------------------------------------------------------- */

int
lookup_channel (char *channel)
{
  int i, nr1, nr2;
  char tag1[5], tag2[5];

  if (isdigit (channel[0]))
    {
      tag1[0] = 0;
      nr1 = atoi (channel);
    }
  else
    {
      sscanf (channel, "%4[A-Za-z]%d", tag1, &nr1);
    }

  for (i = 0; i < CHAN_ENTRIES; i++)
    {
      if (isdigit (tvtuner[i].name[0]))
        {
          tag2[0] = 0;
          nr2 = atoi (tvtuner[i].name);
        }
      else
        {
          sscanf (tvtuner[i].name, "%4[A-Za-z]%d", tag2, &nr2);
        }

      /* First simply test for name equality */
      if (strcmp(channel, tvtuner[i].name) == 0)
        break;

      /* If it doesn't match, try to make a guess */
      if (tag1[0] && tag2[0])
        if (nr1 == nr2 && 0 == strcmp (tag1, tag2))
          break;
      if (!tag1[0] && !tag2[0])
        if (nr1 == nr2)
          break;
    }
  if (i == CHAN_ENTRIES)
    return -1;

  return i;
}

int
get_freq (int i)
{
  if (!tvtuner[i].freq[chan_tab])
    return -1;
  return tvtuner[i].freq[chan_tab] * 16 / 1000;
}

int
cf2freq (char *name, int fine)
{
  int i;

  if (-1 == (i = lookup_channel (name)))
    return -1;
  return get_freq (i) + fine;
}

/* ----------------------------------------------------------------------- */

static struct STRTAB captab[] = {
  {CAPTURE_OFF, "off"},
  {CAPTURE_OFF, "no"},
  {CAPTURE_OFF, "false"},
  {CAPTURE_OVERLAY, "on"},
  {CAPTURE_OVERLAY, "yes"},
  {CAPTURE_OVERLAY, "true"},
  {CAPTURE_OVERLAY, "overlay"},
  {CAPTURE_GRABDISPLAY, "grab"},
  {CAPTURE_GRABDISPLAY, "grabdisplay"},
  {-1, NULL}
};

/* just malloc memory for a new channel ... */
struct CHANNEL *
add_channel (char *name)
{
  static int alloc_count = 0;
  struct CHANNEL *channel;

  if (alloc_count == count)
    {
      alloc_count += 16;
      if (alloc_count == 16)
        channels = malloc (sizeof (struct CHANNEL *) * alloc_count);
      else
        channels =
          realloc (channels, sizeof (struct CHANNEL *) * alloc_count);
    }
  channel = channels[count++] = malloc (sizeof (struct CHANNEL));
  memcpy (channel, &defaults, sizeof (struct CHANNEL));
  channel->name = strdup (name);
  return channel;
}

#ifndef NO_X11
void
hotkey_channel (struct CHANNEL *channel)
{
  char str[100], key[32], ctrl[16];

  if (NULL == channel->key)
    return;
  if (2 == sscanf (channel->key, "%15[A-Za-z0-9_]+%31[A-Za-z0-9_]",
                   ctrl, key))
    sprintf (str, "%s<Key>%s: SetChannel(\"%s\")", ctrl, key, channel->name);
  else
    sprintf (str, "<Key>%s: SetChannel(\"%s\")", channel->key, channel->name);
  XtOverrideTranslations (tv, XtParseTranslationTable (str));
  XtOverrideTranslations (opt_paned, XtParseTranslationTable (str));
  XtOverrideTranslations (chan_viewport, XtParseTranslationTable (str));
}
#endif

/* ... and initalize later */
void
configure_channel (struct CHANNEL *channel)
{
  channel->freq = get_freq (channel->channel) + channel->fine;
#ifndef NO_X11
  channel->button =
    XtVaCreateManagedWidget (channel->name,
                             commandWidgetClass, chan_box,
                             XtNwidth, pix_width,
                             XtNheight, pix_height, NULL);
  XtAddCallback (channel->button, XtNcallback, button_cb,
                 (XtPointer *) channel);
  hotkey_channel (channel);
#endif
}

/* delete channel */
void
del_channel (int i)
{
#ifndef NO_X11
  XtDestroyWidget (channels[i]->button);
#endif
  free (channels[i]->name);
  if (channels[i]->key)
    free (channels[i]->key);
  free (channels[i]);
  count--;
  if (i < count)
    memmove (channels + i, channels + i + 1,
             (count - i) * sizeof (struct CHANNEL *));
}


void
update_channel(struct CHANNEL *dest, struct CHANNEL *src)
{
  Pixmap p;
  Widget b;
  int ck;

  p = dest->pixmap;
  b = dest->button;
  ck = dest->ckey;
  memcpy(dest, src, sizeof(struct CHANNEL));
  dest->pixmap = p;
  dest->button = b;
  dest->ckey = ck;
}

void
debug_channel(struct CHANNEL *chan)
{
  printf("\n");
  printf("channel.name        = %s\n", chan->name);
  printf("channel.key         = %s\n", chan->key);
  printf("channel.cname       = %s\n", chan->cname);
  printf("channel.channel     = %d\n", chan->channel);
  printf("channel.fine        = %d\n", chan->fine);
  printf("channel.freq        = %d\n", chan->freq);
  printf("channel.capture     = %s\n", int_to_str(chan->capture, captab));
  printf("channel.source      = %s\n", int_to_str(chan->source, grabbers[grabber]->inputs));
  printf("channel.norm        = %s\n", int_to_str(chan->norm, grabbers[grabber]->norms));
  printf("channel.color       = %d\n", chan->color);
  printf("channel.bright      = %d\n", chan->bright);
  printf("channel.hue         = %d\n", chan->hue);
  printf("channel.contrast    = %d\n", chan->contrast);
  printf("channel.pixmap      = %ld\n", chan->pixmap);
  printf("channel.button      = %p\n", chan->button);
  printf("channel.ckey        = %d\n", chan->ckey);                
  printf("channel.deinterlace = %d\n", chan->deinterlace);                
  printf("\n");
}

/* ----------------------------------------------------------------------- */
int
read_channel(char *filename, char *line, int nr, struct CHANNEL **pt_current, char *tag, char *val, int flag)
{
  char *h;
  int i;
  struct CHANNEL *current = *pt_current;

  if (1 == sscanf (line, "[%99[^]]]", val))
    {
      if (flag == True)
        {
          current = add_channel (val);
          *pt_current = current;
        }
      else
        current->name = strdup(val);

      //continue;
      return 0;
    }

  if (2 != sscanf (line, " %31[^= ] = %99[^\n]", tag, val))
    {
      fprintf (stderr, "%s:%d: parse error\n", filename, nr);
      //continue;
      return 0;
    }

  for (h = val + strlen (val); h > val && *h == ' '; h--)
    *h = '\0';
  
  if (0 == strcmp (tag, "key"))
    {
      current->key = strdup (val);
    }
  else if (0 == strcmp (tag, "capture"))
    {
      if (-1 != (i = str_to_int (val, captab)))
        current->capture = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else if (0 == strcmp (tag, "source"))
    {
      if (-1 != (i = str_to_int (val, grabbers[grabber]->inputs)))
        current->source = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else if (0 == strcmp (tag, "norm"))
    {
      if (-1 != (i = str_to_int (val, grabbers[grabber]->norms)))
        current->norm = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else if (0 == strcmp (tag, "channel"))
    {
      current->cname = strdup (val);
      current->channel = lookup_channel (current->cname);
      if (-1 == current->channel)
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  // Patch provided by Henk
  //This makes finetuning of freq. configured channels possible
  else if (0 == strcmp(tag, "freq"))
    {
      int n;
      unsigned d;
  
      current->freq = (int)(atof(val) * 1000);
      n = 0;
      d = (unsigned)-1;
      for (i = 0; i < CHAN_ENTRIES; i++)
        {
          if (abs(tvtuner[i].freq[chan_tab] - current->freq) < d)
            {
              d = abs(tvtuner[i].freq[chan_tab] - current->freq);
              n = i;
            }
        }
      current->cname = strdup(tvtuner[n].name);
      current->channel = lookup_channel(current->cname);
      current->fine = (current->freq - tvtuner[n].freq[chan_tab]) * 16 / 1000;
      current->freq = tvtuner[n].freq[chan_tab] * 16 / 1000;
      // debug_channel(current);
    }
  // End of Patch provided by Henk
  else if (0 == strcmp (tag, "fine"))
    {
      current->fine = atoi (val);
    }
  else if (0 == strcmp (tag, "color"))
    {
        current->color = atoi (val);
    }
  else if (0 == strcmp (tag, "bright"))
    {
      current->bright = atoi (val);
    }
  else if (0 == strcmp (tag, "hue"))
    {
      current->hue = atoi (val);
    }
  else if (0 == strcmp (tag, "contrast"))
    {
      current->contrast = atoi (val);
    }
  else if (0 == strcmp (tag, "deinterlace"))
    {
      if (-1 != (i = str_to_int (val, deinterlace_list)))
        current->deinterlace = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else
    {
      // nothing has been find
      return -1;
    }

  // something relevant has been find and hopefully interpreted
  return 0;
}

void
read_config (void)
{
  FILE *fp;
  char filename[100], line[100], tag[32], val[100];
  int i, j, nr = 0;
  float k = 0.0;
  struct CHANNEL *current = &defaults;

  sprintf (filename, "%s/%s", user_dir, "xawdecoderc");
  fp = fopen (filename, "r");
  if (NULL == fp)
    {
      fprintf (stderr, "Can't open config file %s:", filename);
      fprintf (stderr, "writing a default one.\n");
      fprintf (stderr, "Please edit %s to suit your needs\n", filename);
      /* Write a default config file */
      save_config ();

      /* And reread it ! */
      if ((fp = fopen (filename, "r")) == (FILE *) NULL)
        return;
    }
  have_config = 1;

  while (NULL != fgets (line, 99, fp))
    {
      nr++;
      if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
        continue;

      if (read_channel(filename, line, nr, &current, tag, val, True) == 0)
        {
          continue;
        }
      else if (0 == count && 0 == strcmp (tag, "mixer"))
        {
          mixer = strdup (val);
          if (-1 != mixer_open ("/dev/mixer", val))
            have_mixer = 1;
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "freqtab"))
        {
          if (-1 != (i = str_to_int (val, chan_names)))
            chan_tab = i;
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "fullscreen"))
        {
          if (2 != sscanf (val, "%d x %d", &fs_width, &fs_height))
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
              fs_width = fs_height = 0;
            }
        }
      else if (0 == count && 0 == strcmp (tag, "windowsize"))
        {
          if (2 == sscanf (val, "%d x %d", &i, &j))
            {
              video_set_size (i, j, ADAPT_SIZE);
            }
          else
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
            }
        }
      else if (0 == count && 0 == strcmp (tag, "pixsize"))
        {
          if (2 != sscanf (val, "%d x %d", &pix_width, &pix_height))
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
              pix_width = 128;
              pix_height = 96;
            }
        }
      else if (0 == count && 0 == strcmp (tag, "wm-off-by"))
        {
          if (2 != sscanf (val, "%d %d", &fs_xoff, &fs_yoff))
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
              fs_xoff = fs_yoff = 0;
            }
        }
      else if (0 == count && 0 == strcmp (tag, "colorkey"))
        {
          if (1 == sscanf (val, "%d", &i))
            {
              set_colorkey(dpy, i);
            }
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "codec"))
        {
          if (0 == strcmp(val,"DivX"))
            {
              divx.codec = 0;
            }
          else if (0 == strcmp(val,"FFMPEG - Mpeg4"))
            {
              divx.codec = 1;
            }
          else if (0 == strcmp(val,"FFMPEG - Mpeg1Video"))
            {
              divx.codec = 2;
            }
          else if (0 == strcmp(val,"XviD"))
            {
              divx.codec = 3;
            }
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "width"))
        {
          if (1 != sscanf(val,"%d",&divx.width))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "height"))
        {
          if (1 != sscanf(val,"%d",&divx.height))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "bitrate"))
        {
          if (1 != sscanf(val,"%d",&i))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
	  else
	  {
	    divx.bitrate = i*1000;
	  }
        }
     else if (0 == count && 0 == strcmp (tag, "quality"))
       {
         switch(divx.codec)
	   {
	 #ifdef HAVE_DIVX4LINUX
	   case DIVX4LINUX:
	     if ( 0 == strcmp (val, "Fastest"))
	       divx.quality = 1;
	     else if ( 0 == strcmp (val, "Fast"))
	       divx.quality = 2;
	     else if ( 0 == strcmp (val, "Medium"))
	       divx.quality = 3;
	     else if ( 0 == strcmp (val, "Slow"))
	       divx.quality = 4;
	     else if ( 0 == strcmp (val, "Slowest"))
	       divx.quality = 5;
	     else
               fprintf (stderr, "%s:%d: invalid value for %s: %s\n", filename, nr, tag, val);
	     break;
	 #endif
	 #ifdef HAVE_FFMPEG
	   case FFMPEG_MPEG4:
	   case FFMPEG_MPEG1:
	     if ( 0 == strcmp (val, "No ME"))
	       divx.quality = 0;
	     else if ( 0 == strcmp (val, "Very Low"))
	       divx.quality = 1;
	     else if ( 0 == strcmp (val, "Low"))
	       divx.quality = 2;
	     else if ( 0 == strcmp (val, "Medium"))
	       divx.quality = 3;
	     else if ( 0 == strcmp (val, "High"))
	       divx.quality = 4;
	     else if ( 0 == strcmp (val, "Very High"))
	       divx.quality = 5;
	     else
               fprintf (stderr, "%s:%d: invalid value for %s: %s\n", filename, nr, tag, val);
	     break;
	 #endif
	 #ifdef HAVE_XVID
	   case XVID:
	     if ( 0 == strcmp (val, "H623, No ME"))
	       divx.quality = 0;
	     else if ( 0 == strcmp (val, "MPEG, Low"))
	       divx.quality = 1;
	     else if ( 0 == strcmp (val, "H623, Low"))
	       divx.quality = 2;
	     else if ( 0 == strcmp (val, "H623, Medium"))
	       divx.quality = 3;
	     else if ( 0 == strcmp (val, "H623, High"))
	       divx.quality = 4;
	     else if ( 0 == strcmp (val, "Very High"))
	       divx.quality = 5;
	     else if ( 0 == strcmp (val, "Ultra High"))
	       divx.quality = 6;
	     else
               fprintf (stderr, "%s:%d: invalid value for %s: %s\n", filename, nr, tag, val);
	     break;
	 #endif
           }
	}
      else if (0 == count && 0 == strcmp (tag, "stereo_mode"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.stereo = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.stereo = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "compress_audio"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.compress_audio = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.compress_audio = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_bitrate"))
        {
          if (1 != sscanf(val,"%d",&divx.mp3_bitrate))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_quality"))
        {
          if (1 != sscanf(val,"%d",&divx.mp3_quality))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_vbr_mode"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.mp3_vbr = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.mp3_vbr = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_vbr_quality"))
        {
          if (1 != sscanf(val,"%d",&divx.mp3_vbr_quality))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "fps"))
        {
          if (1 != sscanf(val,"%d",&i))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
	  else
	  {
	    divx.fps = i*1000;
	  }
        }
      else if (0 == count && 0 == strcmp (tag, "max_gap"))
        {
          if (1 != sscanf(val,"%f",&k))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
	  else
	  {
	    divx.maxgap = k/1000.0;
	  }
        }
      else if (0 == count && 0 == strcmp (tag, "display_frame"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.display_frames = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.display_frames = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }

      /*
      else if (0 == count && 0 == strcmp (tag, "deinterlace"))
        {
          if (-1 != (i = str_to_int (val, deinterlace_list)))
            {
              set_deinterlace(i);
            }
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      */
      else
        {
          int ret = -1;
          if (plugin != NULL)
            ret = plugin->read_config(&count, tag, val, nr, filename);
          if (ret == -1)
            fprintf (stderr, "%s:%d: unknown tag %s\n", filename, nr, tag);
        }
    }
  fclose (fp);

  /* calculate channel frequencies */
  defaults.channel = lookup_channel (defaults.cname);
  defaults.freq = get_freq (defaults.channel) + defaults.fine;
  for (i = 0; i < count; i++)
    configure_channel (channels[i]);
  create_divxwin();
}

/* ----------------------------------------------------------------------- */

void
save_channel(FILE *fp, int i)
{
  /*
  fprintf (fp, "[%s]\n", channels[i]->name);

  fprintf (fp, "channel = %s\n", tvtuner[channels[i]->channel].name);

  if (cur_fine != channels[i]->fine)
    fprintf (fp, "fine = %+d\n", channels[i]->fine);

  if (cur_norm != channels[i]->norm)
    fprintf (fp, "norm = %s\n",
             int_to_str (channels[i]->norm, grabbers[grabber]->norms));

  if (channels[i]->key != NULL)
    fprintf (fp, "key = %s\n", channels[i]->key);

  if (channels[i]->capture != cur_capture)
    fprintf (fp, "capture = %s\n",
             int_to_str (channels[i]->capture, captab));

  if (channels[i]->source != cur_input)
    fprintf (fp, "source = %s\n",
             int_to_str (channels[i]->source, grabbers[grabber]->inputs));
  
  if (cur_color != channels[i]->color)
    fprintf (fp, "color = %d\n", channels[i]->color);

  if (cur_bright != channels[i]->bright)
    fprintf (fp, "bright = %d\n", channels[i]->bright);

  if (cur_hue != channels[i]->hue)
    fprintf (fp, "hue = %d\n", channels[i]->hue);

  if (cur_contrast != channels[i]->contrast)
    fprintf (fp, "contrast = %d\n", channels[i]->contrast);
  
  fprintf (fp, "\n");
  */

  fprintf (fp, "[%s]\n", channels[i]->name);

  fprintf (fp, "channel = %s\n", tvtuner[channels[i]->channel].name);

  if (defaults.fine != channels[i]->fine)
    fprintf (fp, "fine = %d\n", channels[i]->fine);

  if (defaults.norm != channels[i]->norm)
    fprintf (fp, "norm = %s\n",
             int_to_str (channels[i]->norm, grabbers[grabber]->norms));

  if (channels[i]->key != NULL)
    fprintf (fp, "key = %s\n", channels[i]->key);

  if (defaults.capture != channels[i]->capture)
    fprintf (fp, "capture = %s\n",
             int_to_str (channels[i]->capture, captab));

  if (defaults.source != channels[i]->source)
    fprintf (fp, "source = %s\n",
             int_to_str (channels[i]->source, grabbers[grabber]->inputs));
  
  if (defaults.color != channels[i]->color)
    fprintf (fp, "color = %d\n", channels[i]->color);

  if (defaults.bright != channels[i]->bright)
    fprintf (fp, "bright = %d\n", channels[i]->bright);

  if (defaults.hue != channels[i]->hue)
    fprintf (fp, "hue = %d\n", channels[i]->hue);

  if (defaults.contrast != channels[i]->contrast)
    fprintf (fp, "contrast = %d\n", channels[i]->contrast);
  
  if (defaults.deinterlace != channels[i]->deinterlace)
    fprintf (fp, "deinterlace = %s\n", int_to_str(channels[i]->deinterlace, deinterlace_list));

  fprintf (fp, "\n");
}


void
save_record(FILE *fp)
{
  fprintf (fp, "#\n# record options\n#\n");
  switch(divx.codec)
  {
#ifdef HAVE_DIVX4LINUX
    case DIVX4LINUX:
      fprintf (fp, "codec = DivX\n");
      break;
#endif

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
      fprintf (fp, "codec = FFMPEG - Mpeg4\n");
      break;
    case FFMPEG_MPEG1:
      fprintf (fp, "codec = FFMPEG - Mpeg1Video\n");
      break;
#endif

#ifdef HAVE_XVID
    case XVID:
      fprintf (fp, "codec = XviD\n");
      break;
#endif
   }
   fprintf (fp,"width = %d\n",divx.width);
   fprintf (fp,"height = %d\n",divx.height);
   fprintf (fp,"bitrate = %ld\n",divx.bitrate/1000);
  switch(divx.codec)
  {
#ifdef HAVE_DIVX4LINUX
    case DIVX4LINUX:
      switch(divx.quality)
      {
      	case 1:
      	 fprintf (fp,"quality = Fastest\n");
      	 break;
      	case 2:
      	 fprintf (fp,"quality = Fast\n");
      	 break;
      	case 3:
      	 fprintf (fp,"quality = Medium\n");
      	 break;
      	case 4:
      	 fprintf (fp,"quality = Slow\n");
      	 break;
      	case 5:
      	 fprintf (fp,"quality = Slowest\n");
      	 break;
      }
      break;
#endif

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
    case FFMPEG_MPEG1:
      switch(divx.quality)
      {
      	case 0:
      	 fprintf (fp,"quality = No ME\n");
      	 break;
      	case 1:
      	 fprintf (fp,"quality = Very Low\n");
      	 break;
      	case 2:
      	 fprintf (fp,"quality = Low\n");
      	 break;
      	case 3:
      	 fprintf (fp,"quality = Medium\n");
      	 break;
      	case 4:
      	 fprintf (fp,"quality = High\n");
      	 break;
      	case 5:
      	 fprintf (fp,"quality = Very High\n");
      	 break;
      }
      break;
#endif

#ifdef HAVE_XVID
    case XVID:
      switch(divx.quality)
      {
      	case 0:
      	 fprintf (fp,"quality = H623, No ME\n");
      	 break;
      	case 1:
      	 fprintf (fp,"quality = MPEG, Low\n");
      	 break;
      	case 2:
      	 fprintf (fp,"quality = H623, Low\n");
      	 break;
      	case 3:
      	 fprintf (fp,"quality = H623, Medium\n");
      	 break;
      	case 4:
      	 fprintf (fp,"quality = H623, High\n");
      	 break;
      	case 5:
      	 fprintf (fp,"quality = Very High\n");
      	 break;
      	case 6:
      	 fprintf (fp,"quality = Ultra High\n");
      	 break;
      }
      break;
#endif
   }
   switch(divx.stereo)
   {
     case 0:
       fprintf (fp,"stereo_mode = off\n");
       break;
     case 1:
       fprintf (fp,"stereo_mode = on\n");
       break;
   }
   switch(divx.compress_audio)
   {
     case 0:
       fprintf (fp,"compress_audio = off\n");
       break;
     case 1:
       fprintf (fp,"compress_audio = on\n");
       break;
   }
   fprintf (fp,"mp3_bitrate = %d\n",divx.mp3_bitrate);
   fprintf (fp,"mp3_quality = %d\n",divx.mp3_quality);
   switch(divx.mp3_vbr)
   {
     case 0:
       fprintf (fp,"mp3_vbr_mode = off\n");
       break;
     case 1:
       fprintf (fp,"mp3_vbr_mode = on\n");
       break;
   }
   fprintf (fp,"mp3_vbr_quality = %d\n",divx.mp3_vbr_quality);
   fprintf (fp,"fps = %d\n",divx.fps/1000);
   fprintf (fp,"max_gap = %f\n",divx.maxgap*1000.0);
   switch(divx.display_frames)
   {
     case 0:
       fprintf (fp,"display_frame = off\n");
       break;
     case 1:
       fprintf (fp,"display_frame = on\n");
       break;
   }
   fprintf(fp,"\n");


}

void
save_config (void)
{
  char filename1[100], filename2[100];
  FILE *fp;
  int i;
  Dimension w = 0;
  Dimension h = 0;

  sprintf (filename1, "%s/%s", user_dir, "xawdecoderc");
  sprintf (filename2, "%s/%s", user_dir, "xawdecoderc.bak");

  /* delete old backup */
  unlink (filename2);

  /* current becomes backup */
  if (0 == link (filename1, filename2))
    unlink (filename1);

  /* write new one... */
  fp = fopen (filename1, "w");
  if (NULL == fp)
    {
      fprintf (stderr, "can't open config file %s\n", filename1);
      return;
    }

  /* write defaults */
  fprintf(fp, "#\n# Global options\n#\n");
  /*
  if (cur_norm == -1)
    cur_norm = str_to_int ("SECAM", grabbers[grabber]->norms);
  fprintf (fp, "norm = %s\n",
           int_to_str (cur_norm, grabbers[grabber]->norms));

  if ((cur_capture == -1) || (cur_capture == str_to_int ("off", captab)))
    cur_capture = str_to_int ("grabdisplay", captab);
  fprintf (fp, "capture = %s\n", int_to_str (cur_capture, captab));

  if (cur_input == -1)
    cur_input = str_to_int ("Television", grabbers[grabber]->inputs);
  fprintf (fp, "source = %s\n",
           int_to_str (cur_input, grabbers[grabber]->inputs));

  if ((cur_color != 32768) && (cur_color != 0))
    fprintf (fp, "color = %d\n", cur_color);

  if ((cur_bright != 32768) && (cur_bright != 0))
    fprintf (fp, "bright = %d\n", cur_bright);

  if ((cur_hue != 32768) && (cur_hue != 0))
    fprintf (fp, "hue = %d\n", cur_hue);

  if ((cur_contrast != 32768) && (cur_contrast != 0))
    fprintf (fp, "contrast = %d\n", cur_contrast);
  */

  if (defaults.norm == -1)
    defaults.norm = str_to_int ("SECAM",  grabbers[grabber]->norms);
  fprintf (fp, "norm = %s\n",
           int_to_str (defaults.norm, grabbers[grabber]->norms));

  if ((defaults.capture == -1) || (defaults.capture == str_to_int ("off", captab)))
    defaults.capture = str_to_int ("grabdisplay", captab);
  fprintf (fp, "capture = %s\n", int_to_str (defaults.capture, captab));

  if (defaults.source == -1)
    defaults.source = str_to_int ("Television", grabbers[grabber]->inputs);
  fprintf (fp, "source = %s\n",
           int_to_str (defaults.source, grabbers[grabber]->inputs));

  if ((defaults.color != 32768) && (defaults.color != 0))
    fprintf (fp, "color = %d\n", defaults.color);

  if ((defaults.bright != 32768) && (defaults.bright != 0))
    fprintf (fp, "bright = %d\n", defaults.bright);

  if ((defaults.hue != 32768) && (defaults.hue != 0))
    fprintf (fp, "hue = %d\n", defaults.hue);

  if ((defaults.contrast != 32768) && (defaults.contrast != 0))
    fprintf (fp, "contrast = %d\n", defaults.contrast);

  fprintf (fp, "\n");

  if (fs_width && fs_height)
    fprintf (fp, "fullscreen = %d x %d\n", fs_width, fs_height);
  else
    fprintf (fp, "fullscreen = 768 x 576\n");

  if (fs_xoff || fs_yoff)
    fprintf (fp, "wm-off-by = %+d%+d\n", fs_xoff, fs_yoff);

  fprintf (fp, "freqtab = %s\n", int_to_str (chan_tab, chan_names));
  fprintf (fp, "pixsize = %d x %d\n", pix_width, pix_height);

  XtVaGetValues (tv, XtNwidth, &w, XtNheight, &h, NULL);
  fprintf (fp, "windowsize = %d x %d\n", w, h);

  if (mixer)
    fprintf (fp, "mixer = %s\n", mixer);

  if ((i = get_colorkey(dpy)) != -1)
    fprintf (fp, "colorkey = %d\n", i);

  /*
  fprintf(fp, "# deinterlace: none | bob | weave | greedy | onefield\n"
              "deinterlace = %s\n", int_to_str(deinterlace_method, deinterlace_list));
  */

#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
  save_record(fp);
#endif

  if (plugin != NULL)
    plugin->save_config(fp);

  /* write channels */
  if (count > 0)
    {
      fprintf(fp, "#\n# Channel options\n#\n");
      for (i = 0; i < count; i++)
        {
          save_channel(fp, i);
        }
    }
  else
    {
      fprintf (fp,
               "# For each channel define at least following information\n");
      fprintf (fp, "# [Channel Name]\n");
      fprintf (fp, "# channel = Channel Number\n");
      fprintf (fp, "# norm = PAL|NTSC|SECAM|AUTO\n");
      fprintf (fp,
               "# key = key to press to switch the channel on (KP_End, KP_Next,...)\n");
      fprintf (fp, "# capture = off|grabdisplay|overlay\n");
      fprintf (fp, "\n");
    }



  fclose (fp);
}

int
load_last_channel(struct CHANNEL *last_channel)
{
  FILE *fp;
  char filename[100], line[100], tag[32], val[100];
  int nr = 0;
  int r, status = 0;
  int w, h;

  sprintf (filename, "%s/%s", user_dir, "last_channel");
  fp = fopen (filename, "r");
  if (NULL == fp)
    {
      fprintf(stderr, "load_last_channel: can't open %s\n", filename);
      return -1;
    }

  memcpy(last_channel, &defaults, sizeof(struct CHANNEL));

  while (NULL != fgets (line, 99, fp))
    {
      nr++;
      if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
        continue;


      if ((r=read_channel(filename, line, nr, &last_channel, tag, val, False)) == 0)
        {
          continue;
        }
      else if (0 == strcmp (tag, "windowsize"))
        {
          if (2 == sscanf (val, "%d x %d", &w, &h))
            {
              video_set_size (w, h, ADAPT_SIZE);
            }
          else
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
            }
        }
      else
        {
          // Nothing was found;
          status += r;
        }
    }

  if (nr == 0)
    return -1; // file was empty !!!
  else if (status == 0)
    {
      /* calculate channel frequencies */
      last_channel->freq = get_freq (last_channel->channel) + last_channel->fine;
      hotkey_channel (last_channel);
      return 0; // everything went fine
    }
  else
    return -1;
}

void
save_last_channel(int last_channel)
{
  FILE *fp;
  char filename[100];
  //Dimension w, h;
  
  sprintf (filename, "%s/%s", user_dir, "last_channel");
  fp = fopen (filename, "w");
  if (NULL == fp)
    {
      fprintf(stderr, "save_last_channel: can't open %s\n", filename);
      return;
    }

  // Causes more troubles than real added value
  //XtVaGetValues (tv, XtNwidth, &w, XtNheight, &h, NULL);
  //fprintf (fp, "windowsize = %d x %d\n", w, h);

  save_channel(fp, last_channel);

  fclose(fp);
}


void
save_memcpy_method(int method)
{
  FILE *fp;
  char filename[100];
  
  sprintf (filename, "%s/%s", user_dir, "memcpy_method");
  fp = fopen (filename, "w");

  fprintf(fp, "# memcpy_method: probe | glibc | kernel | mmx | mmxext | sse\n"
              "memcpy_method = %s\n", int_to_str(method, memcpy_list));

  fclose(fp);
}


void
load_memcpy_method(void)
{
  FILE *fp;
  char filename[100];
  char line[100];
  int nr;
  char *h;
  char tag[32], val[100];
  int i;

  sprintf (filename, "%s/%s", user_dir, "memcpy_method");
  fp = fopen (filename, "r");
  if (NULL == fp)
    {
      fprintf(stderr, "load_memcpy_method: can't open %s\n", filename);
      return;
    }

  nr = 0;
  while (NULL != fgets (line, 99, fp))
    {
      nr++;
      if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
        continue;

      if (2 != sscanf (line, " %31[^= ] = %99[^\n]", tag, val))
        {
          fprintf (stderr, "%s:%d: parse error\n", filename, nr);
          continue;
        }

      for (h = val + strlen (val); h > val && *h == ' '; h--)
        *h = '\0';
      

      if (0 == strcmp (tag, "memcpy_method"))
        {
          if (-1 != (i = str_to_int (val, memcpy_list)))
            {
              memcpy_method = i;
            }
          else
            {
              memcpy_method = MEMCPY_PROBE;
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
            }
        }
    }
}

/* ----------------------------------------------------------------------- */

struct STRTAB chan_names[] = {
  {0, "ntsc-bcast"},
  {1, "ntsc-cable"},
  {2, "ntsc-bcast-jp"},
  {3, "ntsc-cable-jp"},
  {4, "secam-france"},
  {5, "pal-europe"},
  {6, "pal-europe-east"},
  {7, "pal-italy"},
  {8, "pal-newzealand"},
  {9, "pal-australia"},
  {10, "pal-ireland"},
  {-1, NULL}
};


/* ------------------------------------------------------------------------- */
/* moved here from channels.h                                                */

/* NOTE : NTSC BROADCAST OVER 69 WERE RE-ALLOCATED CELLULAR, included anyway */

struct freqlist tvtuner[] = {
/* CH    US-TV US-CATV JP-TV JP-CATV SECAMF EUROPE  EUR-E  ITALY    NZ     AU UHF_GHI */
  {"K02", {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"E2", {0, 0, 0, 0, 48250, 48250, 0, 0, 0, 0, 0}},
  {"E3", {0, 0, 0, 0, 55250, 55250, 0, 0, 0, 0, 0}},
  {"E4", {0, 0, 0, 0, 62250, 62250, 0, 0, 0, 0, 0}},

  {"S01", {0, 0, 0, 0, 0, 69250, 0, 0, 0, 0, 0}},
  {"S02", {0, 0, 0, 0, 0, 76250, 0, 0, 0, 0, 0}},
  {"S03", {0, 0, 0, 0, 0, 83250, 0, 0, 0, 0, 0}},

  {"R1", {0, 0, 0, 0, 0, 0, 49750, 0, 0, 0, 0}},
  {"R2", {0, 0, 0, 0, 0, 0, 59250, 0, 0, 0, 0}},
  {"R3", {0, 0, 0, 0, 0, 0, 77250, 0, 0, 0, 0}},
  {"R4", {0, 0, 0, 0, 0, 0, 84250, 0, 0, 0, 0}},
  {"R5", {0, 0, 0, 0, 0, 0, 93250, 0, 0, 0, 0}},

  {"SE1", {0, 0, 0, 0, 105250, 105250, 105250, 0, 0, 0, 0}},
  {"SE2", {0, 0, 0, 0, 112250, 112250, 112250, 0, 0, 0, 0}},
  {"SE3", {0, 0, 0, 0, 119250, 119250, 119250, 0, 0, 0, 0}},
  {"SE4", {0, 0, 0, 0, 126250, 126250, 126250, 0, 0, 0, 0}},
  {"SE5", {0, 0, 0, 0, 133250, 133250, 133250, 0, 0, 0, 0}},
  {"SE6", {0, 0, 0, 0, 140250, 140250, 140250, 0, 0, 0, 0}},
  {"SE7", {0, 0, 0, 0, 147250, 147250, 147250, 0, 0, 0, 0}},
  {"SE8", {0, 0, 0, 0, 154250, 154250, 154250, 0, 0, 0, 0}},
  {"SE9", {0, 0, 0, 0, 161250, 161250, 161250, 0, 0, 0, 0}},
  {"SE10", {0, 0, 0, 0, 168250, 168250, 168250, 0, 0, 0, 0}},

  {"E5", {0, 0, 0, 0, 175250, 175250, 0, 0, 0, 0, 0}},
  {"E6", {0, 0, 0, 0, 182250, 182250, 0, 0, 0, 0, 0}},
  {"E7", {0, 0, 0, 0, 189250, 189250, 0, 0, 0, 0, 0}},
  {"E8", {0, 0, 0, 0, 196250, 196250, 0, 0, 0, 0, 0}},
  {"E9", {0, 0, 0, 0, 203250, 203250, 0, 0, 0, 0, 0}},
  {"E10", {0, 0, 0, 0, 210250, 210250, 0, 0, 0, 0, 0}},
  {"E11", {0, 0, 0, 0, 217250, 217250, 0, 0, 0, 0, 0}},
  {"E12", {0, 0, 0, 0, 224250, 224250, 0, 0, 0, 0, 0}},

  {"R6", {0, 0, 0, 0, 0, 0, 175250, 0, 0, 0, 0}},
  {"R7", {0, 0, 0, 0, 0, 0, 183250, 0, 0, 0, 0}},
  {"R8", {0, 0, 0, 0, 0, 0, 191250, 0, 0, 0, 0}},
  {"R9", {0, 0, 0, 0, 0, 0, 199250, 0, 0, 0, 0}},
  {"R10", {0, 0, 0, 0, 0, 0, 207250, 0, 0, 0, 0}},
  {"R11", {0, 0, 0, 0, 0, 0, 215250, 0, 0, 0, 0}},
  {"R12", {0, 0, 0, 0, 0, 0, 223250, 0, 0, 0, 0}},

  {"SE11", {0, 0, 0, 0, 231250, 231250, 231250, 0, 0, 0, 0}},
  {"SE12", {0, 0, 0, 0, 238250, 238250, 238250, 0, 0, 0, 0}},
  {"SE13", {0, 0, 0, 0, 245250, 245250, 245250, 0, 0, 0, 0}},
  {"SE14", {0, 0, 0, 0, 252250, 252250, 252250, 0, 0, 0, 0}},
  {"SE15", {0, 0, 0, 0, 259250, 259250, 259250, 0, 0, 0, 0}},
  {"SE16", {0, 0, 0, 0, 266250, 266250, 266250, 0, 0, 0, 0}},
  {"SE17", {0, 0, 0, 0, 273250, 273250, 273250, 0, 0, 0, 0}},
  {"SE18", {0, 0, 0, 0, 280250, 280250, 280250, 0, 0, 0, 0}},
  {"SE19", {0, 0, 0, 0, 287250, 287250, 287250, 0, 0, 0, 0}},
  {"SE20", {0, 0, 0, 0, 294250, 294250, 294250, 0, 0, 0, 0}},

  {"S21", {0, 0, 0, 0, 303250, 303250, 303250, 0, 0, 0, 0}},
  {"S22", {0, 0, 0, 0, 311250, 311250, 311250, 0, 0, 0, 0}},
  {"S23", {0, 0, 0, 0, 319250, 319250, 319250, 0, 0, 0, 0}},
  {"S24", {0, 0, 0, 0, 327250, 327250, 327250, 0, 0, 0, 0}},
  {"S25", {0, 0, 0, 0, 335250, 335250, 335250, 0, 0, 0, 0}},
  {"S26", {0, 0, 0, 0, 343250, 343250, 343250, 0, 0, 0, 0}},
  {"S27", {0, 0, 0, 0, 351250, 351250, 351250, 0, 0, 0, 0}},
  {"S28", {0, 0, 0, 0, 359250, 359250, 359250, 0, 0, 0, 0}},
  {"S29", {0, 0, 0, 0, 367250, 367250, 367250, 0, 0, 0, 0}},
  {"S30", {0, 0, 0, 0, 375250, 375250, 375250, 0, 0, 0, 0}},
  {"S31", {0, 0, 0, 0, 383250, 383250, 383250, 0, 0, 0, 0}},
  {"S32", {0, 0, 0, 0, 391250, 391250, 391250, 0, 0, 0, 0}},
  {"S33", {0, 0, 0, 0, 399250, 399250, 399250, 0, 0, 0, 0}},
  {"S34", {0, 0, 0, 0, 407250, 407250, 407250, 0, 0, 0, 0}},
  {"S35", {0, 0, 0, 0, 415250, 415250, 415250, 0, 0, 0, 0}},
  {"S36", {0, 0, 0, 0, 423250, 423250, 423250, 0, 0, 0, 0}},
  {"S37", {0, 0, 0, 0, 431250, 431250, 431250, 0, 0, 0, 0}},
  {"S38", {0, 0, 0, 0, 439250, 439250, 439250, 0, 0, 0, 0}},
  {"S39", {0, 0, 0, 0, 447250, 447250, 447250, 0, 0, 0, 0}},
  {"S40", {0, 0, 0, 0, 455250, 455250, 455250, 0, 0, 0, 0}},
  {"S41", {0, 0, 0, 0, 463250, 463250, 463250, 0, 0, 0, 0}},

  {"K01", {0, 0, 0, 0, 47750, 0, 0, 0, 0, 0, 0}},
  {"K02", {0, 0, 0, 0, 55750, 0, 0, 0, 0, 0, 0}},
  {"K03", {0, 0, 0, 0, 60500, 0, 0, 0, 0, 0, 0}},
  {"K04", {0, 0, 0, 0, 63750, 0, 0, 0, 0, 0, 0}},
  {"K05", {0, 0, 0, 0, 176000, 0, 0, 0, 0, 0, 0}},
  {"K06", {0, 0, 0, 0, 184000, 0, 0, 0, 0, 0, 0}},
  {"K07", {0, 0, 0, 0, 192000, 0, 0, 0, 0, 0, 0}},
  {"K08", {0, 0, 0, 0, 200000, 0, 0, 0, 0, 0, 0}},
  {"K09", {0, 0, 0, 0, 208000, 0, 0, 0, 0, 0, 0}},
  {"K10", {0, 0, 0, 0, 216000, 0, 0, 0, 0, 0, 0}},
  {"K B", {0, 0, 0, 0, 116750, 0, 0, 0, 0, 0, 0}},
  {"K C", {0, 0, 0, 0, 128750, 0, 0, 0, 0, 0, 0}},
  {"K D", {0, 0, 0, 0, 140750, 0, 0, 0, 0, 0, 0}},
  {"K E", {0, 0, 0, 0, 159750, 0, 0, 0, 0, 0, 0}},
  {"K F", {0, 0, 0, 0, 164750, 0, 0, 0, 0, 0, 0}},
  {"K G", {0, 0, 0, 0, 176750, 0, 0, 0, 0, 0, 0}},
  {"K H", {0, 0, 0, 0, 188750, 0, 0, 0, 0, 0, 0}},
  {"K I", {0, 0, 0, 0, 200750, 0, 0, 0, 0, 0, 0}},
  {"K J", {0, 0, 0, 0, 212750, 0, 0, 0, 0, 0, 0}},
  {"K K", {0, 0, 0, 0, 224750, 0, 0, 0, 0, 0, 0}},
  {"K L", {0, 0, 0, 0, 236750, 0, 0, 0, 0, 0, 0}},
  {"K M", {0, 0, 0, 0, 248750, 0, 0, 0, 0, 0, 0}},
  {"K N", {0, 0, 0, 0, 260750, 0, 0, 0, 0, 0, 0}},
  {"K O", {0, 0, 0, 0, 272750, 0, 0, 0, 0, 0, 0}},
  {"K P", {0, 0, 0, 0, 284750, 0, 0, 0, 0, 0, 0}},
  {"K Q", {0, 0, 0, 0, 296750, 0, 0, 0, 0, 0, 0}},
  {"H01", {0, 0, 0, 0, 303250, 0, 0, 0, 0, 0, 0}},
  {"H02", {0, 0, 0, 0, 311250, 0, 0, 0, 0, 0, 0}},
  {"H03", {0, 0, 0, 0, 319250, 0, 0, 0, 0, 0, 0}},
  {"H04", {0, 0, 0, 0, 327250, 0, 0, 0, 0, 0, 0}},
  {"H05", {0, 0, 0, 0, 335250, 0, 0, 0, 0, 0, 0}},
  {"H06", {0, 0, 0, 0, 343250, 0, 0, 0, 0, 0, 0}},
  {"H07", {0, 0, 0, 0, 351250, 0, 0, 0, 0, 0, 0}},
  {"H08", {0, 0, 0, 0, 359250, 0, 0, 0, 0, 0, 0}},
  {"H09", {0, 0, 0, 0, 367250, 0, 0, 0, 0, 0, 0}},
  {"H10", {0, 0, 0, 0, 375250, 0, 0, 0, 0, 0, 0}},
  {"H11", {0, 0, 0, 0, 383250, 0, 0, 0, 0, 0, 0}},
  {"H12", {0, 0, 0, 0, 391250, 0, 0, 0, 0, 0, 0}},
  {"H13", {0, 0, 0, 0, 399250, 0, 0, 0, 0, 0, 0}},
  {"H14", {0, 0, 0, 0, 407250, 0, 0, 0, 0, 0, 0}},
  {"H15", {0, 0, 0, 0, 415250, 0, 0, 0, 0, 0, 0}},
  {"H16", {0, 0, 0, 0, 423250, 0, 0, 0, 0, 0, 0}},
  {"H17", {0, 0, 0, 0, 431250, 0, 0, 0, 0, 0, 0}},
  {"H18", {0, 0, 0, 0, 439250, 0, 0, 0, 0, 0, 0}},
  {"H19", {0, 0, 0, 0, 447250, 0, 0, 0, 0, 0, 0}},

  {"0", {0, 0, 0, 0, 0, 0, 0, 0, 0, 46250, 45750}},
  {"1", {0, 73250, 91250, 0, 0, 0, 0, 0, 45250, 57250, 53750}},
  {"2", {55250, 55250, 97250, 0, 0, 0, 0, 53750, 55250, 64250, 61750}},
  {"3", {61250, 61250, 103250, 0, 0, 0, 0, 62250, 62250, 86250, 175250}},
  {"4", {67250, 67250, 171250, 0, 0, 0, 0, 82250, 175250, 95250, 183250}},
  {"5", {77250, 77250, 177250, 0, 0, 0, 0, 175250, 182250, 102250, 191250}},
  {"5A", {0, 0, 0, 0, 0, 0, 0, 0, 0, 138250, 0}},
  {"6", {83250, 83250, 183250, 0, 0, 0, 0, 183750, 189250, 175250, 199250}},
  {"7", {175250, 175250, 189250, 0, 0, 0, 0, 192250, 196250, 182250, 207250}},
  {"8", {181250, 181250, 193250, 0, 0, 0, 0, 201250, 203250, 189250, 215250}},
  {"9", {187250, 187250, 199250, 0, 0, 0, 0, 210250, 210250, 196250, 0}},
  {"10", {193250, 193250, 205250, 0, 0, 0, 0, 210250, 217250, 209250, 0}},
  {"11", {199250, 199250, 211250, 0, 0, 0, 0, 217250, 0, 216250, 0}},
  {"12", {205250, 205250, 217250, 0, 0, 0, 0, 224250, 0, 0, 0}},

  {"13", {211250, 211250, 0, 109250, 0, 0, 0, 0, 0, 0, 0}},
  {"14", {471250, 121250, 0, 115250, 0, 0, 0, 0, 0, 0, 0}},
  {"15", {477250, 127250, 0, 121250, 0, 0, 0, 0, 0, 0, 0}},
  {"16", {483250, 133250, 0, 127250, 0, 0, 0, 0, 0, 0, 0}},
  {"17", {489250, 139250, 0, 133250, 0, 0, 0, 0, 0, 0, 0}},
  {"18", {495250, 145250, 0, 139250, 0, 0, 0, 0, 0, 0, 0}},
  {"19", {501250, 151250, 0, 145250, 0, 0, 0, 0, 0, 0, 0}},
  {"20", {507250, 157250, 0, 151250, 0, 0, 0, 0, 0, 0, 0}},

  {"21", {513250, 163250, 0, 157250, 471250, 471250, 471250, 0, 471250, 0, 471250}},
  {"22", {519250, 169250, 0, 165250, 479250, 479250, 479250, 0, 479250, 0, 479250}},
  {"23", {525250, 217250, 0, 223250, 487250, 487250, 487250, 0, 487250, 0, 487250}},
  {"24", {531250, 223250, 0, 231250, 495250, 495250, 495250, 0, 495250, 0, 495250}},
  {"25", {537250, 229250, 0, 237250, 503250, 503250, 503250, 0, 503250, 0, 503250}},
  {"26", {543250, 235250, 0, 243250, 511250, 511250, 511250, 0, 511250, 0, 511250}},
  {"27", {549250, 241250, 0, 249250, 519250, 519250, 519250, 0, 519250, 0, 519250}},

  {"28",
   {555250, 247250, 0, 253250, 527250, 527250, 527250, 0, 527250, 527250, 527250}},
  {"29",
   {561250, 253250, 0, 259250, 535250, 535250, 535250, 0, 535250, 534250, 535250}},
  {"30",
   {567250, 259250, 0, 265250, 543250, 543250, 543250, 0, 543250, 541250, 543250}},
  {"31",
   {573250, 265250, 0, 271250, 551250, 551250, 551250, 0, 551250, 548250, 551250}},
  {"32",
   {579250, 271250, 0, 277250, 559250, 559250, 559250, 0, 559250, 555250, 559250}},
  {"33",
   {585250, 277250, 0, 283250, 567250, 567250, 567250, 0, 567250, 562250, 567250}},
  {"34",
   {591250, 283250, 0, 289250, 575250, 575250, 575250, 0, 575250, 569250, 575250}},
  {"35",
   {597250, 289250, 0, 295250, 583250, 583250, 583250, 0, 583250, 576250, 583250}},
  {"36",
   {603250, 295250, 0, 301250, 591250, 591250, 591250, 0, 591250, 0, 591250}},
  {"37",
   {609250, 301250, 0, 307250, 599250, 599250, 599250, 0, 599250, 0, 599250}},
  {"38",
   {615250, 307250, 0, 313250, 607250, 607250, 607250, 0, 607250, 0, 607250}},
  {"39",
   {621250, 313250, 0, 319250, 615250, 615250, 615250, 0, 615250, 604250, 615250}},
  {"40",
   {627250, 319250, 0, 325250, 623250, 623250, 623250, 0, 623250, 611250, 623250}},
  {"41",
   {633250, 325250, 0, 331250, 631250, 631250, 631250, 0, 631250, 618250, 631250}},
  {"42",
   {639250, 331250, 0, 337250, 639250, 639250, 639250, 0, 639250, 625250, 639250}},
  {"43",
   {645250, 337250, 0, 343250, 647250, 647250, 647250, 0, 647250, 632250, 647250}},
  {"44",
   {651250, 343250, 0, 349250, 655250, 655250, 655250, 0, 655250, 639250, 655250}},
  {"45",
   {657250, 349250, 663250, 355250, 663250, 663250, 663250, 0, 663250, 646250,
    663250}},
  {"46",
   {663250, 355250, 669250, 361250, 671250, 671250, 671250, 0, 671250, 653250,
    671250}},
  {"47",
   {669250, 361250, 675250, 367250, 679250, 679250, 679250, 0, 679250, 660250,
    679250}},
  {"48",
   {675250, 367250, 681250, 373250, 687250, 687250, 687250, 0, 687250, 667250,
    687250}},
  {"49",
   {681250, 373250, 687250, 379250, 695250, 695250, 695250, 0, 695250, 674250,
    695250}},
  {"50",
   {687250, 379250, 693250, 385250, 703250, 703250, 703250, 0, 703250, 681250,
    703250}},
  {"51",
   {693250, 385250, 699250, 391250, 711250, 711250, 711250, 0, 711250, 688250,
    711250}},
  {"52",
   {699250, 391250, 705250, 397250, 719250, 719250, 719250, 0, 719250, 695250,
    719250}},
  {"53",
   {705250, 397250, 711250, 403250, 727250, 727250, 727250, 0, 727250, 702250,
    727250}},
  {"54",
   {711250, 403250, 717250, 409250, 735250, 735250, 735250, 0, 735250, 709250,
    735250}},
  {"55",
   {717250, 409250, 723250, 415250, 743250, 743250, 743250, 0, 743250, 716250,
    743250}},
  {"56",
   {723250, 415250, 729250, 421250, 751250, 751250, 751250, 0, 751250, 723250,
    751250}},
  {"57",
   {729250, 421250, 735250, 427250, 759250, 759250, 759250, 0, 759250, 730250,
    759250}},
  {"58",
   {735250, 427250, 741250, 433250, 767250, 767250, 767250, 0, 767250, 737250,
    767250}},
  {"59",
   {741250, 433250, 747250, 439250, 775250, 775250, 775250, 0, 775250, 744250,
    775250}},
  {"60",
   {747250, 439250, 753250, 445250, 783250, 783250, 783250, 0, 783250, 751250,
    783250}},
  {"61",
   {753250, 445250, 759250, 451250, 791250, 791250, 791250, 0, 0, 758250,
    791250}},
  {"62",
   {759250, 451250, 765250, 457250, 799250, 799250, 799250, 0, 799250, 765250,
    799250}},
  {"63",
   {765250, 457250, 0, 463250, 807250, 807250, 807250, 0, 807250, 772250, 807250}},
  {"64",
   {771250, 463250, 0, 0, 815250, 815250, 815250, 0, 815250, 779250, 815250}},
  {"65",
   {777250, 469250, 0, 0, 823250, 823250, 823250, 0, 823250, 786250, 823250}},
  {"66",
   {783250, 475250, 0, 0, 831250, 831250, 831250, 0, 831250, 793250, 831250}},
  {"67",
   {789250, 481250, 0, 0, 839250, 839250, 839250, 0, 839250, 800250, 839250}},
  {"68",
   {795250, 487250, 0, 0, 847250, 847250, 847250, 0, 847250, 807250, 847250}},
  {"69",
   {801250, 493250, 0, 0, 855250, 855250, 855250, 0, 855250, 814250, 855250}},

  {"70", {807250, 499250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"71", {813250, 505250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"72", {819250, 511250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"73", {825250, 517250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"74", {831250, 523250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"75", {837250, 529250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"76", {843250, 535250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"77", {849250, 541250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"78", {855250, 547250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"79", {861250, 553250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"80", {867250, 559250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"81", {873250, 565250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"82", {879250, 571250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"83", {885250, 577250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"84", {0, 583250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"85", {0, 589250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"86", {0, 595250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"87", {0, 601250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"88", {0, 607250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"89", {0, 613250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"90", {0, 619250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"91", {0, 625250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"92", {0, 631250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"93", {0, 637250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"94", {0, 643250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"95", {0, 91250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"96", {0, 97250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"97", {0, 103250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"98", {0, 109250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"99", {0, 115250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"100", {0, 649250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"101", {0, 655250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"102", {0, 661250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"103", {0, 667250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"104", {0, 673250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"105", {0, 679250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"106", {0, 685250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"107", {0, 691250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"108", {0, 697250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"109", {0, 703250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"110", {0, 709250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"111", {0, 715250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"112", {0, 721250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"113", {0, 727250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"114", {0, 733250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"115", {0, 739250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"116", {0, 745250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"117", {0, 751250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"118", {0, 757250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"119", {0, 763250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"120", {0, 769250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"121", {0, 775250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"122", {0, 781250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"123", {0, 787250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"124", {0, 793250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"125", {0, 799250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},

  {"T7", {0, 8250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"T8", {0, 14250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"T9", {0, 20250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"T10", {0, 26250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"T11", {0, 32250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"T12", {0, 38250, 0, 0, 0, 0, 0, 0, 0, 0, 0}},
  {"T13", {0, 44250, 0, 0, 0, 0, 0, 0, 0, 0, 0}}
};
